##modele IDE=group
##Layer=vector
##t=Field Layer
##val=Field Layer
##Covariable=optional Field Layer
##ajouter_lon_lat_comme_covariables=boolean TRUE
##grid_size=number 50
##units=selection secs;mins;hours;days;weeks
##forecast=number 1
##itermax=number 100
##basis=selection  bisquare;Gaussian;exp;Matern32
##nresolution= number 1
##scale_aperture=number 1
##amplitude=selection constant_basis();auto_basis
##aperture=selection constant_basis();auto_basis
##xshift=selection constant_basis();auto_basis
##yshift=selection constant_basis();auto_basis
##effectuer_la_validation_du_modele=boolean TRUE
##pas_de_temps_de_validation=number
##pourcentage_points_aleatoires_de_validation=number 10
##repertoire_resultats=optional folder
##nom_du_fichier= optional string
##Resultat_observations= output vector
library(IDE)
library(dplyr)
library("sp")
library("spacetime")
library("sf")
library("verification")
library("scoringRules")
library("tidyr")
library("ggplot2")
library("gridExtra")
library("FRK")
library(gt)

df<-as.data.frame(Layer)
coords<-st_coordinates(Layer)
crs<-st_crs(Layer)$proj4string
dfc<-cbind(df,coords)
if (!Covariable==""){
data<-dplyr::select(dfc,all_of(t),all_of(val),X,Y,all_of(Covariable))
names(data)<-c("t","cnt","X","Y","Cov")
data$Cov<-as.double(data$Cov)
  }else{
data<-dplyr::select(dfc,t,all_of(val),X,Y)
names(data)<-c("t","cnt","X","Y")
}
data<-drop_na(data)
data$cnt<-as.double(data$cnt)
#conversion des XY en km si pas Lat/lon Nécessaire à cause de bug fit.IDE
convert=FALSE
if (data$Y>100) {
convert=TRUE
data$X<-as.double(data$X/1000)
data$Y<-as.double(data$Y/1000)}
#traitement de la date des données en entrée
is.date <- function(x) inherits(x, 'Date')
is.convertible.to.date <- function(x) !is.na(as.Date(as.character(x),origin="0000-01-01"))
if (!sapply(data$t[1], is.date)&& !is.character(data$t[1])) {
  if (near(data$t[1], as.integer(data$t[1]))){
    data$t<-as.Date(data$t,origin="0000-01-01")
    trans=1
  }
}
if (!sapply(data$t[1], is.date)&& is.character(data$t[1])) {   
  if (sapply(data$t[1], is.convertible.to.date)){
    data$t<-as.Date(data$t)
    trans=2
  }
}
#traitement du pas de temps de validation
if (effectuer_la_validation_du_modele) {
tv<-pas_de_temps_de_validation
if (trans==1) {
tv<-as.Date(tv,origin="0000-01-01")
}else{
tv<-as.Date(tv)
}
t0<-min(data$t)
tmax<-max(data$t)
if (tv<=t0) {
    tv<-t0+1
    }
if (tv>tmax) {
    tv<-tmax-1
} 
if (tv==tmax && forecast==0){
  forecast=1
}  
>paste("pas de temps retenu pour validation=",tv)
}
#Création objet STIDF
spat_part <- SpatialPoints(coords = data[, c("X", "Y")])
temp_part <- data$t
mSTIDF <- STIDF(sp = spat_part,
                time = temp_part,
                data = dplyr::select(data, -t, -X, -Y))
#Création des sousensembles de données de calcul et de validation
valblock_idx=0
valrandom_idx=0
obs_idx=0
if (effectuer_la_validation_du_modele) {
#échantillonnage pour validation
#block pas de temps
mtot<-length(mSTIDF)
mSTIDF$timeY <- format(time(mSTIDF), "%Y-%m-%d")
if (trans==1) {
vt<-as.Date(tv,origin="0000-01-01")
}else{
vt<-as.Date(tv)
}
valblock_idx<-which(mSTIDF$timeY%in%c(as.character(vt)))
obs_idx <- setdiff(1:mtot, valblock_idx)
#pourcentage d'observations
pc<-pourcentage_points_aleatoires_de_validation/100
set.seed(1)
valrandom_idx <- sample(obs_idx,
                        pc * length(obs_idx),
                        replace = FALSE) %>% sort()
obs_idx <- setdiff(obs_idx, valrandom_idx)
obs<-mSTIDF[obs_idx,]
block<-mSTIDF[valblock_idx,]
random<-mSTIDF[valrandom_idx,]
}else{
obs<-mSTIDF
}
>paste("Le bloc de validation du pas de temps comporte",length(valblock_idx),"observations")
>paste("L'&eacute;chantillon al&eacute;atoire de validation comporte",length(valrandom_idx),"observations")
>paste("Le mod&egrave;le sera calcul&eacute; &agrave; partir de",length(obs_idx),"observations")
#fonctions de base
if (basis==0) {base="bisquare"}
if (basis==1) {base="Gaussian"}
if (basis==2) {base="exp"}
if (basis==3) {base="Matern32"}
mbasis_1<-auto_basis(manifold=plane(),
            data= mSTIDF,
            nres=nresolution,
            scale_aperture=scale_aperture,
            type=base)
>"mbasis"
>mbasis_1
>"kernel basis"
kernel_basis<-list()
#if (amplitude==0)  {
  kernel_basis[1]<-list(thetam1=constant_basis())
 # }else {
 # kernel_basis[1]<-list(thetam1=mbasis_1)}
#if (aperture==0 ) {
  kernel_basis[2]<-(thetam2=constant_basis())
#  }else{
#  kernel_basis[2]<-(thetam2=mbasis_1)}
if (xshift==0 ) {
  kernel_basis[3]<-(thetam3= constant_basis())
  }else {
  kernel_basis[3]<-(thetam3= mbasis_1)}
if (yshift==0) {
  kernel_basis[4]<-(thetam4= constant_basis())
  }else {
  kernel_basis[4]<-(thetam4= mbasis_1)}
>kernel_basis
if (units==0) {unit="secs"}
if (units==1) {unit="mins"}
if (units==2) {unit="hours"}
if (units==3) {unit="days"}
if (units==4) {unit="weeks"}
>"calcul du modele IDE..."
#construction de la formule du modèle
form="cnt ~ 1"
if (ajouter_lon_lat_comme_covariables) {
  form<-paste(form,"+ X + Y")
}
if(!Covariable==''){
  form<-paste(form,"+ Cov")
}
form<-as.formula(form)
#construction de l'objet modèle
IDEmodel <- IDE(f = form ,
                data = obs,
                dt = as.difftime(1, units = unit),
                grid_size = grid_size,
                forecast=forecast,
                kernel_basis= kernel_basis)

>"Ajustement du mod&egrave;le..."
fit_results_obs <- fit.IDE(IDEmodel,method="DEoptim",itermax=itermax)
#saveRDS(fit_results_obs,file="c:/STRbook/Book/fit_results_obsCas3H.Rda")
#fit_results_obs<-readRDS("c:/STRbook/Book/fit_results_obsCas3.Rda")
kp<-fit_results_obs$IDEmodel$get("k") %>% unlist()
>"kernel parameters: "
>kp
loglikIDE <- -fit_results_obs$IDEmodel$negloglik()
>"loglikIDE"
>loglikIDE
sigma2_eps<-fit_results_obs$IDEmodel$get("sigma2_eps")
if (repertoire_resultats=='') {
repertoire_resultats=(paste(path.expand('~'),"\\Documents",sep=""))
}
setwd(repertoire_resultats)
sor<-paste("Rplots",nom_du_fichier,".pdf",sep="")
pdf(sor)
sigma2_eps<-fit_results_obs$IDEmodel$get("sigma2_eps")
#sortie tableau coefficients de régression
co<-coef(fit_results_obs$IDEmodel)
codf<-as.data.frame(co)
names(codf)<-c("Valeur")
co_tbl<-as_tibble(codf,rownames="Variable")
tab1<-co_tbl %>%
  gt() %>%
  tab_header(
    title = md("Coefficients de r&eacute;gression des variables du mod&egrave;le")
  ) %>%
  tab_source_note(md("Si les observations ne peuvent pas avoir la valeur 0, l'intercept n'a aucune signification"))
gtsave(tab1,"Coefficient_regression.html")
#sortie tableau paramètres du modèle
ck1<-as.character(kernel_basis[1])
vk1<-strsplit(ck1,"n = ")
par1<-as.numeric(gsub(")","",lapply(vk1,'[[',3)))
ck2<-as.character(kernel_basis[2])
vk2<-strsplit(ck2,"n = ")
par2<-as.numeric(gsub(")","",lapply(vk2,'[[',3)))
ck3<-as.character(kernel_basis[3])
vk3<-strsplit(ck3,"n = ")
par3<-as.numeric(gsub(")","",lapply(vk3,'[[',3)))
ck4<-as.character(kernel_basis[4])
vk4<-strsplit(ck4,"n = ")
par4<-as.numeric(gsub(")","",lapply(vk4,'[[',3)))
teta<-fit_results_obs$IDEmodel$get("k")%>%unlist()
tetadf<-as.data.frame(teta)
names(tetadf)<-c("Valeur")
nr<-length(teta)
tetadf[nr+1,]<-fit_results_obs$IDEmodel$get("sigma2_eps")
tetadf[nr+2,]<-fit_results_obs$IDEmodel$get("sigma2_eta")
rownames(tetadf)[rownames(tetadf) ==nr+1] <- "variance erreur mesure"
rownames(tetadf)[rownames(tetadf) ==nr+2] <- "variance perturbation aleatoire"
cnt=1
for (i in cnt:cnt+par1-1) {
  rownames(tetadf)[rownames(tetadf) ==(paste("par",i,sep=""))] <- paste("Amplitude:",i)
}
cnt=par1+1
for (i in cnt:cnt+par2-1) {
  rownames(tetadf)[rownames(tetadf) ==(paste("par",i,sep=""))] <- paste("Aperture:",i)
}
cnt=par1+par2+1
cnt2=cnt+par3-1
for (i in cnt:cnt2) {
  rownames(tetadf)[rownames(tetadf) ==(paste("par",i,sep=""))] <- paste("xshift:",i-cnt+1)
}
cnt=par1+par2+par3+1
cnt2=cnt+par4-1
for (i in cnt:cnt2) {
  rownames(tetadf)[rownames(tetadf) ==(paste("par",i,sep=""))] <- paste("yshift:",i-cnt+1)
}
teta_tbl<-as_tibble(tetadf,rownames="Variable")
tab2<-teta_tbl %>%
  gt() %>%
  tab_header(
    title = md("**Param&egrave;tres calcul&eacute;s du mod&egrave;le**")
  ) %>%
  tab_source_note(md("*Les premiers param&egrave;tres correspondent aux valeurs de &theta;.Les deux derniers correspondent &agrave; : - la variance des observations et &agrave; -la perturbation ( param&egrave;tre &eta;t). *")) %>%
  tab_source_note(
    source_note = md("Le nombre d'items par param&egrave;tre &theta; d&eacute;pend si le param&egrave;tre est variant ou pas")
  )
gtsave(tab2,"Parametres_modele.html")
#sortie table eigenvalues M
abs_ev <- eigen(fit_results_obs$IDEmodel$get("M"))$values %>%abs()
abse<-as.matrix(summary(abs_ev))
colnames(abse) <- c("Valeur")
abs_tbl<-as.data.frame(abse)
abs_tbl<-as_tibble(abs_tbl,rownames="Parametre")
tab3<-abs_tbl %>%
  gt() %>%
  tab_header(
    title = md("**Valeurs propres de la matrice d'&eacute;volution M**")
  ) %>%
  tab_source_note(md("*Si la valeur la plus grande est inf&eacute;rieure &agrave; 1, le processus est stable*"))
gtsave(tab3,"M_eigenvalues.html")
>"R&eacute;alisation des pr&eacute;dictions..."
>"Realisation des predictions..."
>"Grille complete"
if(Covariable==''){
>"Grille complete"
ST_grid <- predict(fit_results_obs$IDEmodel)
ST_grid_df <- ST_grid %>%
  data.frame() %>%
  mutate(IDE_pred = ST_grid$Ypred,
         IDE_predse = ST_grid$Ypredse,
         IDE_predZse = sqrt(ST_grid$Ypredse^2 +
                              sigma2_eps))
ST_grid_df$t<-as.integer(format(ST_grid_df$t,"%d"))-1
if(convert) {
ST_grid_df$X<-ST_grid_df$X*1000
ST_grid_df$Y<-ST_grid_df$Y*1000}
coords<-dplyr::select(ST_grid_df,X,Y)
attr<-dplyr::select(ST_grid_df,-cnt,-X,-Y,-IDE_pred,-IDE_predse)
names(attr)<-c("t","Predict","ErreurY","ErreurZ")
spdfN <- SpatialPointsDataFrame(coords = coords, data = attr,proj4string=CRS(crs))
#Resultat_grille<-st_as_sf(spdfN)
sfn<-st_as_sf(spdfN)
nomfic<-paste(repertoire_resultats,"/",nom_du_fichier,"_grille.shp",sep="")
st_write(sfn, nomfic, driver="ESRI Shapefile",append=FALSE)
}

>"Previsions observations"
pred_IDE_obs <- predict(fit_results_obs$IDEmodel,
                          newdata = obs)
obs_df <- obs %>%
  data.frame() %>%
  mutate(IDE_pred = pred_IDE_obs$Ypred,
         IDE_predse = pred_IDE_obs$Ypredse,
         IDE_predZse = sqrt(IDE_predse^2 +
                              sigma2_eps))
obs_df$t<-as.integer(format(obs_df$time,"%d"))-1
if(convert) {
obs_df$X<-obs_df$X*1000
obs_df$Y<-obs_df$Y*1000}
coords<-dplyr::select(obs_df,X,Y)
attr<-dplyr::select(obs_df,sp.ID,t,cnt,IDE_pred,IDE_predse,IDE_predZse)
names(attr)<-c("ID","t","Val","Predict","ErreurY","ErreurZ")
attr$t<-as.integer(attr$t)
spdfN <- SpatialPointsDataFrame(coords = coords, data = attr,proj4string=CRS(crs))
Resultat_observations<-st_as_sf(spdfN)

p1<-show_kernel(fit_results_obs$IDEmodel,scale=0.2)
plot(p1)

c1<-cor(obs_df$cnt,obs_df$IDE_pred, method="pearson")
titre<-paste("Observations-Predictions vs observations. Correlation=",c1 )
p1<-ggplot(obs_df,aes(cnt, IDE_pred)) + 
  ggtitle(titre) + geom_point() + 
  stat_smooth(method="lm", se=FALSE)
plot(p1)

if (effectuer_la_validation_du_modele) {

>"Predictions des donnees de validation"
#prédictions pour block pas de temps
pred_IDE_block <- predict(fit_results_obs$IDEmodel,
                          newdata = block)
block_df <- block %>%
  data.frame() %>%
  mutate(IDE_pred = pred_IDE_block$Ypred,
         IDE_predse = pred_IDE_block$Ypredse,
         IDE_predZse = sqrt(IDE_predse^2 +
                              sigma2_eps))
block_df$t<-as.integer(format(block_df$time,"%d"))-1
if(convert) {
block_df$X<-block_df$X*1000
block_df$Y<-block_df$Y*1000}
coords<-dplyr::select(block_df,X,Y)
attr<-dplyr::select(block_df,sp.ID,t,cnt,IDE_pred,IDE_predse,IDE_predZse)
names(attr)<-c("ID","t","Val","Predict","ErreurY","ErreurZ")
attr$t<-as.integer(attr$t)
spdfN <- SpatialPointsDataFrame(coords = coords, data = attr,proj4string=CRS(crs))
sfn<-st_as_sf(spdfN)
nomfic<-paste(repertoire_resultats,"/",nom_du_fichier,"_block.shp",sep="")
st_write(sfn, nomfic, driver="ESRI Shapefile",append=FALSE)
#prédictions pour pourcentage échantillons
pred_IDE_random <- predict(fit_results_obs$IDEmodel,
                           newdata = random)
alea_df <- random %>%
  data.frame() %>%
 mutate(IDE_pred = pred_IDE_random$Ypred,
         IDE_predse = pred_IDE_random$Ypredse,
         IDE_predZse = sqrt(IDE_predse^2 +
                             sigma2_eps))
alea_df$t<-as.integer(format(alea_df$time,"%d"))-1
if(convert) {
alea_df$X<-alea_df$X*1000
alea_df$Y<-alea_df$Y*1000}
coords<-dplyr::select(alea_df,X,Y)
attr<-dplyr::select(alea_df,sp.ID,t,cnt,IDE_pred,IDE_predse,IDE_predZse)
names(attr)<-c("ID","t","Val","Predict","ErreurY","ErreurZ")
attr$t<-as.integer(attr$t)
spdfN <- SpatialPointsDataFrame(coords = coords, data = attr,proj4string=CRS(crs))
sfn<-st_as_sf(spdfN)
nomfic<-paste(repertoire_resultats,"/",nom_du_fichier,"_alea.shp",sep="")
st_write(sfn, nomfic, driver="ESRI Shapefile",append=FALSE)

c1<-cor(block_df$cnt,block_df$IDE_pred, method="pearson")
titre<-paste("Bloc-Predictions vs observations. Correlation=",c1 )
p2<-ggplot(block_df,aes(cnt, IDE_pred)) + 
  ggtitle(titre) + geom_point() + 
  stat_smooth(method="lm", se=FALSE)
plot(p2)
c2<-cor(alea_df$cnt,alea_df$IDE_pred, method="pearson")
titre<-paste("Aleatoire-Predictions vs observations. Correlation=",c2 )
p3<-ggplot(alea_df,aes(cnt, IDE_pred)) + 
  ggtitle(titre) + geom_point() + 
  stat_smooth(method="lm", se=FALSE)
plot(p3)
Bias <- function(x,y) mean(x - y) # x: val. obs.
PCV <- function(x,y) mean((x - y)^2) # y: predictions
SCV <- function(x,y,v) mean((x - y)^2 / v) # v: pred. variances
CRPS <- function(x, y, s) verification::crps(x, cbind(y, s))$CRPS
Diagblock <- block_df %>% summarise(
  Bias_bloc = Bias(IDE_pred, cnt),
  PCV_bloc = PCV(IDE_pred, cnt),
  SCV_bloc = SCV(IDE_pred, cnt, IDE_predZse^2),
  CRPS_bloc = CRPS(cnt, IDE_pred, IDE_predZse)
)
Diagrandom <- alea_df %>% summarise(
  Bias_alea = Bias(IDE_pred, cnt),
  PCV_alea = PCV(IDE_pred, cnt),
  SCV_alea = SCV(IDE_pred, cnt, IDE_predZse^2),
  CRPS_alea = CRPS(cnt, IDE_pred, IDE_predZse)
)
>"Diagnostics de la validation par bloc temporel"
>Diagblock
>"Diagnostics de la validation par echantillonnage aleatoire"
>Diagrandom
nval<-length(block)
pred_IDE_block <- predict(fit_results_obs$IDEmodel,
            newdata = block,
            covariances = TRUE)
Veps <- diag(rep(sigma2_eps, nval))
L_IDE <- t(chol(pred_IDE_block$Cov + Veps))
IntIDE <- coef(fit_results_obs$IDEmodel)
nsim <- 100
E <- matrix(rnorm(nval*nsim), nval, nsim)
Sims_IDE <- IntIDE + pred_IDE_block$newdata$Ypred + L_IDE %*% E
es1<-es_sample(block$cnt, dat = as.matrix(Sims_IDE))
>paste("Valeur de es_sample bloc=",es1)
distances <- block %>%
  coordinates() %>%
  dist() %>%
  as.matrix()
weights <- 0.5^distances
vs1<-vs_sample(block$cnt, dat = as.matrix(Sims_IDE),
          w = weights, p = 1)
>paste("Valeur de vs_sample bloc=",vs1)

nval<-length(random)
pred_IDE_random <- predict(fit_results_obs$IDEmodel,
            newdata = random,
            covariances = TRUE)
Veps <- diag(rep(sigma2_eps, nval))
L_IDE <- t(chol(pred_IDE_random$Cov + Veps))
IntIDE <- coef(fit_results_obs$IDEmodel)
E <- matrix(rnorm(nval*nsim), nval, nsim)
Sims_IDE <- IntIDE + pred_IDE_random$newdata$Ypred + L_IDE %*% E
es2<-es_sample(random$cnt, dat = as.matrix(Sims_IDE))
>paste("Valeur de es_sample aléa=",es2)
distances <- random %>%
  coordinates() %>%
  dist() %>%
  as.matrix()
weights <- 0.5^distances
vs2<-vs_sample(random$cnt, dat = as.matrix(Sims_IDE),
          w = weights, p = 1)
>paste("Valeur de vs_sample aléa=",vs2)
loglikIDE <- -fit_results_obs$IDEmodel$negloglik()
pIDE <- 3 +length(kp)
m<-length(obs)
Criteria <- data.frame(AIC = c(0), BIC = c(0),
                       row.names = c( nom_du_fichier))
aic<- -2*loglikIDE + 2*pIDE
bic<- -2*loglikIDE + pIDE*log(m)
Criteria[nom_du_fichier, "AIC"] <- aic
Criteria[nom_du_fichier, "BIC"] <- bic
>Criteria
plot.new()
grid.table(Criteria)
Bias_bloc = Bias(block_df$IDE_pred, block_df$cnt)
PCV_bloc = PCV(block_df$IDE_pred, block_df$cnt)
SCV_bloc = SCV(block_df$IDE_pred, block_df$cnt, block_df$IDE_predZse^2)
CRPS_bloc = CRPS(block_df$cnt, block_df$IDE_pred, block_df$IDE_predZse)
Bias_alea = Bias(alea_df$IDE_pred, alea_df$cnt)
PCV_alea = PCV(alea_df$IDE_pred, alea_df$cnt)
SCV_alea = SCV(alea_df$IDE_pred, alea_df$cnt, alea_df$IDE_predZse^2)
CRPS_alea = CRPS(alea_df$cnt, alea_df$IDE_pred, alea_df$IDE_predZse)
resume <- matrix(c("Correlation",c1,c2,"Biais",Bias_bloc,Bias_alea,"PCV",PCV_bloc,PCV_alea,"SCV",SCV_bloc,SCV_alea,"CRPS",CRPS_bloc,CRPS_alea,"ES",es1,es2,"VS",vs1,vs2),ncol=3,byrow=TRUE)
colnames(resume) <- c("Score","Bloc t","Aleatoire")
resume <- as.table(resume)
>resume
plot.new()
grid.table(resume,rows=NULL)
}
